import json
import os

class MissingConfigFile(Exception): pass

########################
###   DEFAULT
########################

class Default(object):
    """
        Wrapper for default values
        So we can tell the difference between default and no value
    """
    def __init__(self, val):
        self.val = val

    def __str__(self):
        return str(self.val)

    def __unicode__(self):
        return unicode(self.val)

    def append(self, val):
        """Make sure argparse doesn't fail"""
        self.val.append(val)

########################
###   CONFIGUTIL
########################

class ConfigUtil(object):
    """Config functionality"""
    def __init__(self, template=None):
        self.values = {}
        self.template = template

    def normalise_key(self, key):
        """Make sure key is a valid python attribute"""
        key = key.replace('-', '_')
        if key.startswith("noy_"):
            key = key[4:]
        return key

    def find_value(self, key):
        """Find a value and return it"""
        values = self.values
        if key not in values:
            raise AttributeError("Config has no value for {}".format(key))

        val = values[key]
        if isinstance(val, Default):
            return val.val
        else:
            return val

    def use_options(self, options, extractor=None):
        """
            If extractor isn't specified, then just update self.values with options.

            Otherwise update values with whatever the result of calling extractor with
            our template and these options returns

            Also make sure all keys are transformed into valid python attribute names
        """
        # Extract if necessary
        if not extractor:
            extracted = options
        else:
            extracted = extractor(self.template, options)

        # Get values as [(key, val), ...]
        if isinstance(extracted, dict):
            extracted = extracted.items()

        # Add our values if there are any
        # Normalising the keys as we go along
        if extracted is not None:
            for key, val in extracted:
                self.values[self.normalise_key(key)] = val

    def find_config_file(self):
        """
            Find where our config file is if there is any

            If the value for the config file is a default and it doesn't exist
            then it is silently ignored.

            If however, the value isn't a default and it doesn't exist, an error is raised
        """
        filename = self.values.get('config_file', Default('noy.json'))

        ignore_missing = False
        if isinstance(filename, Default):
            filename = filename.val
            ignore_missing = True

        filename = os.path.abspath(filename)
        if os.path.exists(filename):
            return filename
        elif not ignore_missing:
            raise MissingConfigFile("Config file doesn't exist at {}".format(filename))

    def apply_config_file(self, filename):
        """
            Add options from config file to self.values
            Leave alone existing values that are not an instance of Default
        """
        def extractor(template, options):
            """Ignore things that are existing non default values"""
            for name, val in options:
                normalised = self.normalise_key(name)
                if normalised in self.values and not isinstance(self.values[normalised], Default):
                    continue
                else:
                    yield name, val

        items = json.load(open(filename)).items()
        self.use_options(items, extractor)

    def use_config_file(self):
        """Find and apply the config file"""
        self.config_file = self.find_config_file()
        if self.config_file:
            self.apply_config_file(self.config_file)

########################
###   CONFIG
########################

class Config(object):
    """
        A wrapper around ConfigUtil.

        This provides a setup function that knows how to initialise from some options
        And how to find a config file and apply those values
        (config file values are overwritten by the options passed into setup)

        It then provides a __getattr__ to access the values on the ConfigUtil.
    """
    def __init__(self, template=None):
        self._util = ConfigUtil(template)

    def setup(self, options=None, extractor=None):
        """
            Put options onto the config and put anything from a config file onto the config.

            If extractor is specified, it is used to extract values from the options dictionary
        """
        # Get our programmatic options
        self._util.use_options(options, extractor)

        # Overwrite non defaults in self.values with values from config
        self._util.use_config_file()

    def __getattr__(self, key):
        """Get us attributes from the class or from self.values"""
        if key.startswith("_") or key in self.__dict__:
            return object.__getattribute__(self, key)

        return self._util.find_value(key)




from noseOfYeti.tokeniser.config import Default
import os

def default_from_env(str, dflt=None, as_list=False):
    def get(env):
        arg_to_use = env.get('NOSE_NOY_EXTRA_IMPORTS')
        if arg_to_use is None:
            arg_to_use = dflt

        if as_list:
            if arg_to_use:
                return [arg_to_use]
            else:
                return []
        else:
            return dflt
    return get

spec_options = {
      'with-default-imports' : dict(
          default = default_from_env('NOSE_NOY_WITH_DEFAULT_IMPORTS', dflt=False)
        , action  = 'store_true'
        , dest    = 'with_default_imports'
        , help    = 'Turn on default imports for spec files'
        , type    = 'yn'
        )

    , 'no-describe-attrs' : dict(
          default = default_from_env('NOSE_NOY_NO_DESCRIBE_ATTRS', dflt=False)
        , action  = 'store_true'
        , dest    = 'no_describe_attrs'
        , help    = 'Turn off giving describes a is_noy_spec attribute'
        , type    = 'yn'
        )

    , 'default-kls' : dict(
          default = default_from_env('NOSE_NOY_DEFAULT_KLS', dflt='object')
        , action  = 'store'
        , dest    = 'default_kls'
        , help    = 'Set default class for describes'
        , type    = 'string'
        )

    , 'extra-import' : dict(
          default = default_from_env('NOSE_NOY_EXTRA_IMPORTS', as_list=True)
        , action  = 'append'
        , dest    = 'extra_import'
        , help    = '''Set extra default imports
                    (i.e. 'from something import *'
                          'import thing')
                    '''
        , type    = 'csv'
        )

    , 'wrapped-setup' : dict(
          default = default_from_env('NOSE_NOY_WRAPPED_SETUP', dflt=False)
        , action = 'store_true'
        , dest = 'wrapped_setup'
        , help = '''Wrap setups with lines at bottom of file instead of using noy_sup_* helpers'''
        , type = 'yn'
        )

    , 'config-file' : dict(
          default = default_from_env('NOSE_NOY_CONFIG_FILE', dflt='noy.json')
        , dest = 'config_file'
        , help = '''Location of a config file for nose of yeti'''
        )
    }

def extract_options_dict(template, options):
    """Extract options from a dictionary against the template"""
    for option, val in template.items():
        if options and option in options:
            yield option, options[option]
        else:
            yield option, Default(template[option]['default'](os.environ))




'''
Tests to make sure registering the spec codec actually works
'''
from should_dsl import should
import subprocess
import sys
import os

from .helpers import a_temp_file, a_temp_dir

# Silencing code checker about should_dsl matchers
be = None
equal_to = None
start_with = None

init_codez = """
from noseOfYeti.tokeniser.spec_codec import register_from_options
register_from_options()
"""

example_specd_tests = """
# coding: spec

describe "blah":
    it "should totally work":
        1 + 1 |should| equal_to(2)


if hasattr(TestBlah, 'test_should_totally_work'):
    print("test_should_totally_work")
""".strip()

class Test_RegisteringCodec(object):
    def test_not_registering_codec_leads_to_error(self):
        with a_temp_file(example_specd_tests) as filename:
            process = subprocess.Popen([sys.executable, filename], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            process.wait()
            process.returncode |should| be(1)

            expected_output = 'File "{}", line 1\nSyntaxError: encoding problem'.format(filename)
            process.stdout.read().decode('utf8').strip() |should| start_with(expected_output)

    def test_registering_codec_doesnt_lead_to_error(self):
        with a_temp_dir() as tempdir:
            with open(os.path.join(tempdir, "codez.py"), 'w') as fle:
                fle.write(example_specd_tests)

            with open(os.path.join(tempdir, "spec_setup.py"), 'w') as fle:
                fle.write(init_codez)

            filename = os.path.join(tempdir, "script.py")
            with open(filename, 'w') as fle:
                fle.write("import spec_setup; import codez")

            process = subprocess.Popen([sys.executable, filename], cwd=tempdir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            process.wait()
            if process.returncode != 0:
                print(process.stdout.read().decode('utf8'))
            process.returncode |should| be(0)

            expected_output = 'test_should_totally_work'
            process.stdout.read().decode('utf8').strip() |should| equal_to(expected_output)



from tokenize import untokenize
from should_dsl import matcher
from textwrap import dedent
from six import StringIO

@matcher
class ResultIn(object):

    name = 'result_in'

    def __call__(self, radicand):
        self._radicand = radicand
        return self

    def match(self, actual):
        tokeniser, self._actual = actual
        self._actual = dedent(self._actual).strip()
        s = StringIO(self._actual)
        try:
            tokens = tokeniser.translate(s.readline)
            self._expected = untokenize(tokens)
        finally:
            s.close()

        self._expected = dedent(self._expected).strip()
        self._radicand = dedent(self._radicand).strip()

        normalised_expected = []
        for line in self._expected.split('\n'):
            normalised_expected.append(line.rstrip())
        self._expected = '\n'.join(normalised_expected)

        return self._expected == self._radicand

    def message_for_failed_should(self):
        return 'expected "{0}"\n======================>\n"{1}"\n\n======================$\n"{2}"'.format(
            *(res.replace(' ', '.').replace('\t', '-') for res in (self._actual, self._radicand, self._expected)))

    def message_for_failed_should_not(self):
        return 'expected "{0}"\n\tTo not translate to "{1}"'.format(
            *(res.replace(' ', '.').replace('\t', '-') for res in (self._actual, self._radicand)))




from noseOfYeti.tokeniser.tokens import tokens_in

from tokenize import OP, NEWLINE
import six

def determine_imports(with_default_imports=False, extra_imports=None):
    default = []

    if extra_imports:
        if isinstance(extra_imports, six.string_types):
            default.extend(tokens_in(extra_imports))
        else:
            for d in extra_imports:
                if d[0] == NEWLINE:
                    # I want to make sure the extra imports don't
                    # Take up extra lines in the code than the "# coding: spec"
                    default.append((OP, ';'))
                else:
                    default.append(d)

    if with_default_imports:
        if default and tuple(default[-1]) != (OP, ';'):
            default.append((OP, ';'))

        import_tokens = tokens_in('import nose; from nose.tools import *; from noseOfYeti.tokeniser.support import *')
        default.extend(import_tokens)

    return default




from noseOfYeti.tokeniser import Tokeniser, determine_imports

from should_dsl import should, should_not
from tokenize import NEWLINE

# Silencing code checker about should_dsl matchers
contain = None
equal_to = None
result_in = None

class Test_Tokeniser(object):

    def test_gives_describes_noy_specific_attributes(self):
        imports = determine_imports(with_default_imports=False)
        tok = Tokeniser(import_tokens = imports)
        (tok, 'describe "Something testable"') |should| result_in(
        '''
        class TestSomethingTestable (object ):pass

        TestSomethingTestable .is_noy_spec =True
        '''
        )

    def test_is_possible_to_turn_off_attributes(self):
        imports = determine_imports(with_default_imports=False)
        tok = Tokeniser(import_tokens=imports, with_describe_attrs=False)
        (tok, 'describe "Something testable"') |should| result_in('class TestSomethingTestable (object ):pass')

    def test_no_newline_in_default_imports(self):
        tok = Tokeniser(import_tokens=determine_imports())
        tok.import_tokens |should_not| contain([NEWLINE, '\n'])

    def test_no_newline_in_extended_default_imports(self):
        imports = determine_imports(extra_imports='import another.class', with_default_imports=True)
        tok = Tokeniser(import_tokens=imports)
        tok.import_tokens |should_not| contain([NEWLINE, '\n'])
        (tok, '') |should| result_in(
            'import another .class ;import nose ;from nose .tools import *;from noseOfYeti .tokeniser .support import *'
        )

    def test_tokeniser_has_no_default_imports_by_default(self):
        tok = Tokeniser()
        tok.import_tokens |should| equal_to(None)

    def test_is_possible_to_specify_extra_imports_without_default_imports(self):
        imports = determine_imports(with_default_imports=False, extra_imports="import thing")
        tok = Tokeniser(import_tokens = imports)
        (tok, '') |should| result_in('import thing ')

    def test_determine_imports_imports_nothing_by_default(self):
        imports = determine_imports()
        tok = Tokeniser(import_tokens=imports)
        (tok, '') |should| result_in('')




from should_dsl import should
import fudge

from noseOfYeti.tokeniser.config import Default, Config
from .helpers import a_temp_file

import six

# Silencing code checker about should_dsl matchers
be = None
equal_to = None

class Test_Default(object):

    def test_stringify_returns_value(self):
        str(Default("blah")) |should| equal_to("blah")
        if six.PY2:
            unicode(Default("blah")) |should| equal_to("blah")

        str(Default(1)) |should| equal_to("1")
        if six.PY2:
            unicode(Default(1)) |should| equal_to("1")

        class Special(object):
            def __unicode__(self):
                return 'things'
            def __str__(self):
                return 'stuff'
        str(Default(Special())) |should| equal_to("stuff")
        if six.PY2:
            unicode(Default(Special())) |should| equal_to("things")

class Test_ConfigSetup(object):
    '''Test that the Config class knows how to setup itself'''
    def test_it_updates_values_with_options(self):
        # With no values to begin with
        config = Config()
        config._util.values |should| equal_to({})
        config.setup({'a':1, 'b':2})
        config._util.values |should| equal_to({'a':1, 'b':2})

    def test_it_uses_extractor_if_provided(self):
        '''If given an extractor, it applies extractor to the options before setting them'''
        template = fudge.Fake("template")
        def extractor(template, options):
            for key, val in options.items():
                yield key, val + 1

        config = Config(template)
        config.setup({'a':1, 'b':2}, extractor=extractor)
        config._util.values |should| equal_to({'a':2, 'b':3})

    def test_it_gets_values_from_config_file(self):
        '''It gets values from a config file'''
        contents = """
            { "a" : 1
            , "b" : 2
            }
        """
        with a_temp_file(contents) as filename:
            config = Config()
            config.setup({"config-file":filename})
            config._util.values |should| equal_to({"config_file":filename, u"a":1, u"b":2})

    def test_it_overwrites_config_file_with_options(self):
        '''It gets values from a config file'''
        contents = """
            { "a" : 1
            , "b" : 2
            }
        """
        with a_temp_file(contents) as filename:
            config = Config()
            config.setup({"config-file":filename, "a":4})
            config._util.values |should| equal_to({"config_file":filename, "a":4, u"b":2})

    @fudge.patch("noseOfYeti.tokeniser.config.ConfigUtil")
    def test_setup_uses_util(self, FakeConfigUtil):
        '''It uses self._util in setup'''
        options = fudge.Fake("options")
        template = fudge.Fake("template")
        extractor = fudge.Fake("extractor")
        util = (fudge.Fake("util").remember_order()
            .expects("use_options").with_args(options, extractor)
            .expects("use_config_file")
            )

        FakeConfigUtil.expects_call().with_args(template).returns(util)
        config = Config(template)
        config.setup(options, extractor=extractor)

class Test_ConfigAttributes(object):
    '''Make sure you can get values from config like they were attributes'''

    def test_config_gets_values_like_attributes(self):
        '''Make sure the options we set can be retrieved like values'''
        config = Config()
        config.setup({"a":1, "b":2, "c-d":3})

        config.a |should| be(1)
        config.b |should| be(2)
        config.c_d |should| be(3)

    @fudge.test
    def test_config_gets_attributes_from_util(self):
        '''__getattr__ on Config uses values from util'''
        key = 'my_awesome_key'
        val = fudge.Fake("val")

        util = fudge.Fake("util").expects("find_value").with_args(key).returns(val)
        config = Config()
        config._util = util

        getattr(config, key) |should| be(val)




from inspect import getmembers

class TestChooser(object):
    def __init__(self):
        self.new_module()

    def new_module(self):
        """Tells TestChooser that a new module has been entered"""
        self.done = {}

    def already_visited(self, kls, name):
        """Determine if a method has already been accepted for this module"""
        key = '%s.%s' % (kls, name)
        if key not in self.done:
            self.done[key] = True
            return False
        else:
            return True

    def consider(self, method, ignore_kls=None):
        """
            Determines whether a method should be considered a Test
            Returns False if it believes it isn't a test
            Will return True otherwise

            ignore_kls should be a list of classes to ignore
        """
        if not ignore_kls:
            ignore_kls = []

        if method.__name__.startswith("ignore__"):
            # Method wants to be ignored
            return False

        if hasattr(method, '__test__') and not method.__test__:
            # Method doesn't want to be tested
            return False

        kls = None
        if getattr(method, "__self__"):
            kls = method.__self__.__class__

        if not kls:
            # im_class seems to be None in pypy
            for k, v in getmembers(method):
                if k == 'im_self' and v:
                    kls = v.__class__
                    break
                elif k == 'im_class' and v:
                    kls = v
                    break

        if kls.__name__ in ignore_kls:
            # Kls should be ignored
            return False

        if not hasattr(kls, 'is_noy_spec'):
            # Kls not a noy_spec, we don't care if it runs or not
            return None

        if kls.__dict__.get("__only_run_tests_in_children__"):
            # Only run these tests in the children, not in this class itself
            return False

        method_in_kls = method.__name__ in kls.__dict__
        method_is_test = method.__name__.startswith('test_')
        method_passed_down = any(
            method.__name__ in superkls.__dict__ and getattr(superkls, "__only_run_tests_in_children__", False)
            for superkls in kls.__bases__
        )

        if (method_passed_down or method_in_kls) and method_is_test:
            if not self.already_visited(kls.__name__, method.__name__):
                return True

        # Is a noy_spec method but not a valid test, refuse it
        return False




from noseOfYeti.tokeniser import Tokeniser
from should_dsl import should

import six

# Silencing code checker about should_dsl matchers
result_in = None

func_accessor = ""
if six.PY2:
    func_accessor = "__func__ ."

class Test_Tokenisor_translation(object):
    def setUp(self):
        self.toka = Tokeniser(with_describe_attrs=False)
        self.tokb = Tokeniser(with_describe_attrs=False, default_kls = 'other')

    def test_translates_a_describe(self):
        (self.toka, 'describe "Something testable"') |should| result_in('class TestSomethingTestable (object ):pass')
        (self.tokb, 'describe "Something testable"') |should| result_in('class TestSomethingTestable (other ):pass')

        # Same tests, but with newlines in front
        (self.toka, '\ndescribe "Something testable"') |should| result_in('\nclass TestSomethingTestable (object ):pass')
        (self.tokb, '\ndescribe "Something testable"') |should| result_in('\nclass TestSomethingTestable (other ):pass')

    def test_translates_an_it(self):
        (self.toka, 'it "should do this thing":') |should| result_in('def test_should_do_this_thing ():')
        (self.tokb, 'it "should do this thing":') |should| result_in('def test_should_do_this_thing ():')

        # Same tests, but with newlines in front
        (self.toka, '\nit "should do this thing":') |should| result_in('\ndef test_should_do_this_thing ():')
        (self.tokb, '\nit "should do this thing":') |should| result_in('\ndef test_should_do_this_thing ():')

    def test_adds_arguments_to_its_if_declared_on_same_line(self):
        (self.toka, 'it "should do this thing", blah, meh:') |should| result_in(
            'def test_should_do_this_thing (blah ,meh ):'
            )
        (self.tokb, 'it "should do this thing", blah, meh:') |should| result_in(
            'def test_should_do_this_thing (blah ,meh ):'
            )

        # Same tests, but with newlines in front
        (self.toka, '\nit "should do this thing", blah, meh:') |should| result_in(
            '\ndef test_should_do_this_thing (blah ,meh ):'
            )
        (self.tokb, '\nit "should do this thing", blah, meh:') |should| result_in(
            '\ndef test_should_do_this_thing (blah ,meh ):'
            )

    def test_adds_arguments_to_its_if_declared_on_same_line_and_work_with_skipTest(self):
        (self.toka, 'it "should do this thing", blah, meh') |should| result_in(
            'def test_should_do_this_thing (blah ,meh ):raise nose.SkipTest '
            )
        (self.tokb, 'it "should do this thing", blah, meh') |should| result_in(
            'def test_should_do_this_thing (blah ,meh ):raise nose.SkipTest '
            )

        # Same tests, but with newlines in front
        (self.toka, '\nit "should do this thing", blah, meh') |should| result_in(
            '\ndef test_should_do_this_thing (blah ,meh ):raise nose.SkipTest '
            )
        (self.tokb, '\nit "should do this thing", blah, meh') |should| result_in(
            '\ndef test_should_do_this_thing (blah ,meh ):raise nose.SkipTest '
            )

    def test_no_added_arguments_to_its_if_not_declared_on_same_line(self):
        (self.toka, 'it "should do this thing"\n, blah, meh') |should| result_in(
            "def test_should_do_this_thing ():raise nose.SkipTest\n,blah ,meh "
            )
        (self.tokb, 'it "should do this thing"\n, blah, meh') |should| result_in(
            "def test_should_do_this_thing ():raise nose.SkipTest\n,blah ,meh "
            )

        # Same tests, but with newlines in front
        (self.toka, '\nit "should do this thing"\n, blah, meh') |should| result_in(
            "\ndef test_should_do_this_thing ():raise nose.SkipTest\n,blah ,meh "
            )
        (self.tokb, '\nit "should do this thing"\n, blah, meh') |should| result_in(
            "\ndef test_should_do_this_thing ():raise nose.SkipTest\n,blah ,meh "
            )

    def test_turns_an_it_without_colon_into_skippable(self):
        (self.toka, 'it "should be skipped"\n') |should| result_in(
            'def test_should_be_skipped ():raise nose.SkipTest '
        )

        (self.toka, 'it "should not be skipped":\n') |should| result_in(
            'def test_should_not_be_skipped ():'
        )

        # Same tests, but with newlines in front
        (self.toka, '\nit "should be skipped"\n') |should| result_in(
            '\ndef test_should_be_skipped ():raise nose.SkipTest '
        )

        (self.toka, '\nit "should not be skipped":\n') |should| result_in(
            '\ndef test_should_not_be_skipped ():'
        )

    def test_turns_before_each_into_setup(self):
        (self.toka, 'before_each:') |should| result_in('def setUp (self ):')

        # Same tests, but with newlines in front
        (self.toka, '\nbefore_each:') |should| result_in('\ndef setUp (self ):')

    def test_indentation_should_work_regardless_of_crazy_groups(self):
        test = """
        describe 'a':
            it 'asdf':
                l = [ True
                    , False
                        , 1
                , 2
                ]

                t = (1
                        , 2
                , 3
            , 4
                    , 5,
                )

                d = {'asdf' : True}

                t2 = (True
                ,    False
        )
            it 'asdf2'"""

        desired = """
        class TestA (object ):
            def test_asdf (self ):
                l =[True
                ,False
                ,1
                ,2
                ]

                t =(1
                ,2
                ,3
                ,4
                ,5 ,
                )

                d ={'asdf':True }

                t2 =(True
                ,False
                )
            def test_asdf2 (self ):raise nose.SkipTest
        """

        (self.toka, test) |should| result_in(desired)

    def test_indentation_for_test_should_work_after_skipped_test(self):
        test = """
        describe 'thing':
            it 'should be skipped'
            it 'shouldnt be skipped':
                print 'hi'

            it 'another that should be skipped'

            it 'another that shouldnt be skipped':
                print 'hi'"""

        desired = """
        class TestThing (object ):
            def test_should_be_skipped (self ):raise nose.SkipTest
            def test_shouldnt_be_skipped (self ):
                print 'hi'

            def test_another_that_should_be_skipped (self ):raise nose.SkipTest

            def test_another_that_shouldnt_be_skipped (self ):
                print 'hi'
        """

        (self.toka, test) |should| result_in(desired)

    def test_indentation_for_describe_should_work_after_skipped_test(self):
        test = '''
        describe 'thing':
            it 'should be skipped'
            describe 'that':
                pass'''

        desired = '''
        class TestThing (object ):
            def test_should_be_skipped (self ):raise nose.SkipTest
        class TestThing_That (TestThing ):
            pass
        '''
        (self.toka, test) |should| result_in(desired)

    def test_indentation_should_work_for_inline_python_code(self):
        test = '''
        describe 'this':
            describe 'that':
                pass

        class SomeMockObject(object):
            def indented_method()
        '''

        desired = '''
        class TestThis (object ):pass
        class TestThis_That (TestThis ):
            pass

        class SomeMockObject (object ):
            def indented_method ()
        '''

        (self.toka, test) | should | result_in(desired)

    def test_gives_setups_super_call_when_in_describes(self):
        test = '''
        describe "Thing":
            before_each:
                self.x = 5
        '''

        desired = '''
        class TestThing (object ):
            def setUp (self ):
                noy_sup_setUp (super (TestThing ,self ));self .x =5
        '''

        (self.toka, test) |should| result_in(desired)
        # and with tabs
        (self.toka, test.replace('    ', '\t')) |should| result_in(desired.replace('    ', '\t'))

    def test_turns_after_each_into_teardown(self):
        (self.toka, 'after_each:') |should| result_in('def tearDown (self ):')

        # Same tests, but with newlines in front
        (self.toka, '\nafter_each:') |should| result_in('\ndef tearDown (self ):')

    def test_gives_teardowns_super_call_when_in_describes(self):
        test = '''
        describe "Thing":
            after_each:
                self.x = 5
        '''

        desired = '''
        class TestThing (object ):
            def tearDown (self ):
                noy_sup_tearDown (super (TestThing ,self ));self .x =5
        '''

        (self.toka, test) |should| result_in(desired)
        # and with tabs
        (self.toka, test.replace('    ', '\t')) |should| result_in(desired.replace('    ', '\t'))

    def test_it_has_the_ability_to_wrap_setup_and_tearDown_instead_of_inserting_sup_call(self):
        self.toka.wrapped_setup = True
        self.tokb.wrapped_setup = True

        test = '''
        describe "Thing":
            after_each:
                self.x = 5

            describe "Other":
                before_each:
                    self.y = 8
        '''

        desired = '''
        class TestThing (object ):
            def tearDown (self ):
                self .x =5

        class TestThing_Other (TestThing ):
            def setUp (self ):
                self .y =8

        TestThing .tearDown =noy_wrap_tearDown (TestThing ,TestThing .tearDown )
        TestThing_Other .setUp =noy_wrap_setUp (TestThing_Other ,TestThing_Other .setUp )
        '''

        (self.toka, test) |should| result_in(desired)
        (self.toka, test.replace('    ', '\t')) |should| result_in(desired.replace('    ', '\t'))

    def test_it_is_possible_to_have_indented_block_after_setup_with_wrapped_setup_option(self):
        self.toka.wrapped_setup = True
        self.tokb.wrapped_setup = True

        test = '''
        describe "Thing":
            after_each:
                def blah(self):
                    pass

            describe "Other":
                before_each:
                    class Stuff(object):
                        pass
        '''

        desired = '''
        class TestThing (object ):
            def tearDown (self ):
                def blah (self ):
                    pass

        class TestThing_Other (TestThing ):
            def setUp (self ):
                class Stuff (object ):
                    pass

        TestThing .tearDown =noy_wrap_tearDown (TestThing ,TestThing .tearDown )
        TestThing_Other .setUp =noy_wrap_setUp (TestThing_Other ,TestThing_Other .setUp )
        '''

        (self.toka, test) |should| result_in(desired)
        (self.toka, test.replace('    ', '\t')) |should| result_in(desired.replace('    ', '\t'))


    def test_has_ignorable_its(self):
        (self.toka, '\nignore "should be ignored"') |should| result_in(
            '\ndef ignore__should_be_ignored ():raise nose.SkipTest '
            )
        (self.toka, '\nignore "should be ignored"') |should| result_in(
            '\ndef ignore__should_be_ignored ():raise nose.SkipTest '
            )

    def test_no_transform_inside_expression(self):
        (self.toka, 'variable = before_each') |should| result_in('variable =before_each ')
        (self.toka, 'variable = after_each')  |should| result_in('variable =after_each ')
        (self.toka, 'variable = describe')    |should| result_in('variable =describe ')
        (self.toka, 'variable = ignore')      |should| result_in('variable =ignore ')
        (self.toka, 'variable = it')          |should| result_in('variable =it ')

        # Same tests, but with newlines in front
        (self.toka, '\nvariable = before_each') |should| result_in('\nvariable =before_each ')
        (self.toka, '\nvariable = after_each')  |should| result_in('\nvariable =after_each ')
        (self.toka, '\nvariable = describe')    |should| result_in('\nvariable =describe ')
        (self.toka, '\nvariable = ignore')      |should| result_in('\nvariable =ignore ')
        (self.toka, '\nvariable = it')          |should| result_in('\nvariable =it ')

    def test_allows_definition_of_different_base_class_for_next_describe(self):
        test = '''
        describe unittest.TestCase "This thing":pass
        describe "Another thing":pass
        '''

        desired = '''%s
        class TestThisThing (unittest .TestCase ):pass
        class TestAnotherThing (%s ):pass
        '''

        (self.toka, test) |should| result_in(desired % ('', 'object'))
        (self.tokb, test) |should| result_in(desired % ('', 'other'))

        # Same tests, but with newlines in front
        (self.toka, '\n%s' % test) |should| result_in(desired % ('\n', 'object'))
        (self.tokb, '\n%s' % test) |should| result_in(desired % ('\n', 'other'))

    def test_sets__testname__on_non_alphanumeric_test_names(self):
        test = '''
        it "(root level) should work {well}"
            3 |should| be(4)
        describe "SomeTests":
            it "doesn't get phased by $special characters"

            describe "NestedDescribe":
                it "asdf $% asdf":
                    1 |should| be(2)
        it "(root level) should also [work]"
        '''

        desired = '''
        def test_root_level_should_work_well ():raise nose.SkipTest
            3 |should |be (4 )
        class TestSomeTests (%s ):
            def test_doesnt_get_phased_by_special_characters (self ):raise nose.SkipTest

        class TestSomeTests_NestedDescribe (TestSomeTests ):
            def test_asdf_asdf (self ):
                1 |should |be (2 )
        def test_root_level_should_also_work ():raise nose.SkipTest
        test_root_level_should_work_well .__testname__ ="(root level) should work {{well}}"
        test_root_level_should_also_work .__testname__ ="(root level) should also [work]"
        TestSomeTests .test_doesnt_get_phased_by_special_characters .{func_accessor}__testname__ ="doesn't get phased by $special characters"
        TestSomeTests_NestedDescribe .test_asdf_asdf .{func_accessor}__testname__ ="asdf $%% asdf"
        '''.format(func_accessor=func_accessor)

        (self.toka, test) |should| result_in(desired % "object")
        (self.tokb, test) |should| result_in(desired % "other")

    def test_it_maintains_line_numbers_when_pass_on_another_line(self):
        test = '''
        it "is a function with a pass": pass

        it "is a function with a pass on another line":
            pass

        it "is a function with a pass on another line further below":
            #comment or something


            pass

        describe "block with a pass":
            pass

        describe "block with comment and pass":
            # comment or something
            pass

        describe "Nesting and passes": pass
            # comment
            describe "Nested":
                pass

                describe "More Nesting":
                    # comment


                    pass
        '''

        desired = '''
        def test_is_a_function_with_a_pass ():pass

        def test_is_a_function_with_a_pass_on_another_line ():
            pass

        def test_is_a_function_with_a_pass_on_another_line_further_below ():
        #comment or something


            pass

        class TestBlockWithAPass (%(dflt)s ):
            pass

        class TestBlockWithCommentAndPass (%(dflt)s ):
        # comment or something
            pass

        class TestNestingAndPasses (%(dflt)s ):pass
        # comment
        class TestNestingAndPasses_Nested (TestNestingAndPasses ):
            pass

        class TestNestingAndPasses_Nested_MoreNesting (TestNestingAndPasses_Nested ):
        # comment


            pass
        '''

        (self.toka, test) |should| result_in(desired % {'dflt': "object"})
        (self.tokb, test) |should| result_in(desired % {'dflt': "other"})

    def test_it_allows_default_arguments_for_its(self):
        test = '''
        it "is a test with default arguments", thing=2, other=[3]

        describe "group":
            it "has self and default args", blah=None, you=(3, 4,
                5, 5):
                # Test space is respected

                1 |should| be(2)
        '''

        desired = '''
        def test_is_a_test_with_default_arguments (thing =2 ,other =[3 ]):raise nose.SkipTest

        class TestGroup (%(dflt)s ):
            def test_has_self_and_default_args (self ,blah =None ,you =(3 ,4 ,
            5 ,5 )):
            # Test space is respected

                1 |should |be (2 )
        '''

        (self.toka, test) |should| result_in(desired % {'dflt': "object"})
        (self.tokb, test) |should| result_in(desired % {'dflt': "other"})

    def test_can_properly_dedent_after_block_of_just_containers(self):
        test = '''
        it "should ensure askers are None or boolean or string":
            for val in (None, False, 'asdf', u'asdf', lambda: 1):
                (lambda : Step(askBeforeAction  = val)) |should_not| throw(Problem)
                (lambda : Step(askDesiredResult = val)) |should_not| throw(Problem)
                (lambda : Step(blockBeforeGet   = val)) |should_not| throw(Problem)

            for val in (1, True):
                (lambda : Step(askBeforeAction  = val)) |should| throw(Problem)
                (lambda : Step(askDesiredResult = val)) |should| throw(Problem)
                (lambda : Step(blockBeforeGet   = val)) |should| throw(Problem)

            3 |should| be(3)
        '''

        desired = '''
        def test_should_ensure_askers_are_None_or_boolean_or_string ():
            for val in (None ,False ,'asdf',u'asdf',lambda :1 ):
                (lambda :Step (askBeforeAction =val ))|should_not |throw (Problem )
                (lambda :Step (askDesiredResult =val ))|should_not |throw (Problem )
                (lambda :Step (blockBeforeGet =val ))|should_not |throw (Problem )

            for val in (1 ,True ):
                (lambda :Step (askBeforeAction =val ))|should |throw (Problem )
                (lambda :Step (askDesiredResult =val ))|should |throw (Problem )
                (lambda :Step (blockBeforeGet =val ))|should |throw (Problem )

            3 |should |be (3 )
        '''

        (self.toka, test) |should| result_in(desired)
        (self.tokb, test) |should| result_in(desired)

    def test_it_doesnt_add_semicolon_after_noy_setup_if_not_necessary(self):
        test = '''
            describe "block with necessary semicolon":
                before_each:
                    two = 1 + 1

            describe "block with unecessary semiclon":
                before_each:
                    #comment
                    pass

                after_each:

                    pass
        '''

        desired = '''
        class TestBlockWithNecessarySemicolon (%(dflt)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestBlockWithNecessarySemicolon ,self ));two =1 +1

        class TestBlockWithUnecessarySemiclon (%(dflt)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestBlockWithUnecessarySemiclon ,self ))#comment
                pass

            def tearDown (self ):
                noy_sup_tearDown (super (TestBlockWithUnecessarySemiclon ,self ))
                pass
        '''
        (self.toka, test) |should| result_in(desired % {'dflt': "object"})
        (self.tokb, test) |should| result_in(desired % {'dflt': "other"})

    def test_it_keeps_comments_placed_after_setup_and_teardown_methods(self):
        test = '''
            describe "Kls":
                before_each: # Comment one

                    pass

                after_each: # Comment two

                    pass

            describe "Kls2":
                before_each: # Comment three
                    two = 1 + 1

                after_each: # Comment four
                    #comment
                    pass
        '''

        desired = '''
        class TestKls (%(dflt)s ):
            def setUp (self ):# Comment one
                noy_sup_setUp (super (TestKls ,self ))
                pass

            def tearDown (self ):# Comment two
                noy_sup_tearDown (super (TestKls ,self ))
                pass

        class TestKls2 (%(dflt)s ):
            def setUp (self ):# Comment three
                noy_sup_setUp (super (TestKls2 ,self ));two =1 +1

            def tearDown (self ):# Comment four
                noy_sup_tearDown (super (TestKls2 ,self ))#comment
                pass
        '''
        (self.toka, test) |should| result_in(desired % {'dflt': "object"})
        (self.tokb, test) |should| result_in(desired % {'dflt': "other"})




from tokenize import NAME, OP
import re

regexes = {
      'joins': re.compile('[- /]')
    , 'punctuation': re.compile('[+\-*/=\$%^&\'",.:;?{()}#<>\[\]]')
    , 'repeated_underscore': re.compile('_{2,}')
    }

def acceptable(value, capitalize=False):
    """Convert a string into something that can be used as a valid python variable name"""
    name = regexes['punctuation'].sub("", regexes['joins'].sub("_", value))
    # Clean up irregularities in underscores.
    name = regexes['repeated_underscore'].sub("_", name.strip('_'))
    if capitalize:
        # We don't use python's built in capitalize method here because it
        # turns all upper chars into lower chars if not at the start of
        # the string and we only want to change the first character.
        name_parts = []
        for word in name.split('_'):
            name_parts.append(word[0].upper())
            if len(word) > 1:
                name_parts.append(word[1:])
        name = ''.join(name_parts)
    return name

class TokenDetails(object):
    """Container for current token"""
    def __init__(self, tokenum=None, value=None, scol=0):
        self.set(tokenum, value, scol)

    def set(self, tokenum, value, scol):
        self.scol = scol
        self.value = value
        self.tokenum = tokenum

    def transfer(self, details):
        details.set(*self.values())

    def values(self):
        return self.tokenum, self.value, self.scol

class Single(object):
    """Container for a single block (i.e. it or ignore block)"""
    def __init__(self, group, typ=None, indent=0):
        self.typ = typ
        self.group = group
        self.indent = indent

        self.args = []
        self._name = None
        self.english = None
        self.skipped = False

        self.starting_arg = False

        if not self.group.root:
            self.args.append([(NAME, 'self')])

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = acceptable(value)
        self.english = value

    @property
    def python_name(self):
        if self.typ == 'it':
            return "test_%s" % self.name
        else:
            return "ignore__%s" % self.name

    @property
    def identifier(self):
        if self.group.root:
            return self.python_name
        else:
            return "%s.%s" % (self.group.kls_name, self.python_name)

    def add_to_arg(self, tokenum, value):
        if tokenum == OP and value ==',':
            self.starting_arg = True
            return

        if self.starting_arg:
            self.args.append([(tokenum, value)])
            self.starting_arg = False
        else:
            self.args[-1].append((tokenum, value))

class Group(object):
    """Container for group blocks (i.e. describe or context)"""
    def __init__(self, name=None, root=False, parent=None, level=0, typ=None):
        self.kls = None
        self.typ = typ
        self.name = name
        self.root = root
        self.empty = True
        self.level = level
        self.parent = parent
        self.singles = []
        self.has_after_each = False
        self.has_before_each = False

        # Default whether this group is starting anything
        self.starting_single = False
        if root:
            # Root technically isn't a group, so it doesn't have a signature to start
            self.starting_group = False
        else:
            # Group is created before we have all information
            # Hence it's signature is being created
            self.starting_group = True

    def __repr__(self):
        if self.root:
            return "<Root Group>"
        else:
            return "<Group %s:%s(%s)>" % (self.kls_name, self.super_kls, self.parent)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.english = value
        if value:
            self._name = acceptable(value, True)

    @property
    def starting_signature(self):
        """Determine if this group is starting itself or anything belonging to it"""
        return self.starting_group or self.starting_single

    @property
    def kls_name(self):
        """Determine python name for group"""
        # Determine kls for group
        if not self.parent or not self.parent.name:
            return 'Test{0}'.format(self.name)
        else:
            use = self.parent.kls_name
            if use.startswith('Test'):
                use = use[4:]

            return 'Test{0}_{1}'.format(use, self.name)

    @property
    def super_kls(self):
        """
            Determine what kls this group inherits from
            If default kls should be used, then None is returned
        """
        if not self.kls and self.parent and self.parent.name:
            return self.parent.kls_name
        return self.kls

    def start_group(self, scol, typ):
        """Start a new group"""
        return Group(parent=self, level=scol, typ=typ)

    def start_single(self, typ, scol):
        """Start a new single"""
        self.starting_single = True
        single = self.single = Single(typ=typ, group=self, indent=(scol - self.level))
        self.singles.append(single)
        return single

    def finish_signature(self):
        """Tell group it isn't starting anything anymore"""
        self.starting_group = False
        self.starting_single = False

    def modify_kls(self, name):
        """Add a part to what will end up being the kls' superclass"""
        if self.kls is None:
            self.kls = name
        else:
            self.kls += name




from noseOfYeti.tokeniser.tracker import Tracker
from noseOfYeti.tokeniser.tokens import Tokens

from tokenize import generate_tokens

class Tokeniser(object):
    """Endpoint for tokenising a file"""
    def __init__(self, default_kls='object', with_describe_attrs=True, import_tokens=None, wrapped_setup=False):
        self.default_kls = default_kls
        self.import_tokens = import_tokens
        self.wrapped_setup = wrapped_setup
        self.with_describe_attrs = with_describe_attrs

    ########################
    ###   TRANSLATE
    ########################

    def translate(self, readline, result=None, no_imports=None):
        # Tracker to keep track of information as the file is processed
        self.tokens = Tokens(self.default_kls)
        self.tracker = Tracker(result, self.tokens, self.wrapped_setup)

        # Add import stuff at the top of the file
        if self.import_tokens and no_imports is not True:
            self.tracker.add_tokens(self.import_tokens)

        # Looking at all the tokens
        with self.tracker.add_phase() as tracker:
            for tokenum, value, (_, scol), _, _ in generate_tokens(readline):
                tracker.next_token(tokenum, value, scol)

        # Add attributes to our Describes so that the plugin can handle some nesting issues
        # Where we have tests in upper level describes being run in lower level describes
        if self.with_describe_attrs:
            self.tracker.add_tokens(self.tracker.make_describe_attrs())

        # If setups should be wrapped, then do this at the bottom
        if self.wrapped_setup:
            self.tracker.add_tokens(self.tracker.wrapped_setups())

        # Add lines to bottom of file to add __testname__ attributes
        self.tracker.add_tokens(self.tracker.make_method_names())

        # Return translated list of tokens
        return self.tracker.result




from tokenize import NAME, OP, INDENT, NEWLINE, DEDENT, STRING, ERRORTOKEN, COMMENT
from contextlib import contextmanager
import re

from noseOfYeti.tokeniser.containers import TokenDetails, Group

# Regex for matching whitespace
regexes = {'whitespace': re.compile('\s+')}

class WildCard(object):
    """Used to determine if tokens should be inserted untill ignored token"""
    def __repr__(self):
        return "<WildCard>"

class Tracker(object):
    """Keep track of what each next token should mean"""
    def __init__(self, result, tokens, wrapped_setup):
        if result is None:
            self.result = []
        else:
            self.result = result

        self.single = None
        self.tokens = tokens
        self.groups = Group(root=True)
        self.current = TokenDetails()
        self.all_groups = [self.groups]
        self.in_container = False
        self.wrapped_setup = wrapped_setup

        self.containers = []
        self.ignore_next = []
        self.indent_amounts = []
        self.adjust_indent_at = []

        self.indent_type = ' '
        self.insert_till = None
        self.after_space = True
        self.inserted_line = False
        self.just_ended_container = False
        self.just_started_container = False

    @contextmanager
    def add_phase(self):
        """Context manager for when adding all the tokens"""
        # add stuff
        yield self

        # Make sure we output eveything
        self.finish_hanging()

        # Remove trailing indents and dedents
        while len(self.result) > 1 and self.result[-2][0] in (INDENT, ERRORTOKEN, NEWLINE):
            self.result.pop(-2)

    def next_token(self, tokenum, value, scol):
        """Determine what to do with the next token"""

        # Make self.current reflect these values
        self.current.set(tokenum, value, scol)

        # Determine indent_type based on this token
        if self.current.tokenum == INDENT and self.current.value:
            self.indent_type = self.current.value[0]

        # Only proceed if we shouldn't ignore this token
        if not self.ignore_token():
            # Determining if this token is whitespace
            self.determine_if_whitespace()

            # Determine if inside a container
            self.determine_inside_container()

            # Change indentation as necessary
            self.determine_indentation()

            # See if we are force inserting this token
            if self.forced_insert():
                return

            # If we have a newline after an inserted line, then we don't need to worry about semicolons
            if self.inserted_line and self.current.tokenum == NEWLINE:
                self.inserted_line = False

            # If we have a non space, non comment after an inserted line, then insert a semicolon
            if self.result and not self.is_space and self.inserted_line:
                if self.current.tokenum != COMMENT:
                    self.result.append((OP, ';'))
                self.inserted_line = False

            # Progress the tracker
            self.progress()

            # Add a newline if we just skipped a single
            if self.single and self.single.skipped:
                self.single.skipped = False
                self.result.append((NEWLINE, '\n'))

            # Set after_space so next line knows if it is after space
            self.after_space = self.is_space

    ########################
    ###   PROGRESS
    ########################

    def progress(self):
        """
            Deal with next token
            Used to create, fillout and end groups and singles
            As well as just append everything else
        """
        tokenum, value, scol = self.current.values()

        # Default to not appending anything
        just_append = False

        # Prevent from group having automatic pass given to it
        # If it already has a pass
        if tokenum == NAME and value == 'pass':
            self.groups.empty = False

        # Set variables to be used later on to determine if this will likely make group not empty
        created_group = False
        found_content = False
        if not self.groups.starting_group and not self.is_space:
            found_content = True

        if self.groups.starting_group:
            # Inside a group signature, add to it
            if tokenum == STRING:
                self.groups.name = value

            elif tokenum == NAME or (tokenum == OP and value == '.'):
                # Modify super class for group
                self.groups.modify_kls(value)

            elif tokenum == NEWLINE:
                # Premature end of group
                self.add_tokens_for_group(with_pass=True)

            elif tokenum == OP and value == ":":
                # Proper end of group
                self.add_tokens_for_group()

        elif self.groups.starting_single:
            # Inside single signature, add to it
            if tokenum == STRING:
                self.single.name = value

            elif tokenum == NEWLINE and not self.in_container:
                # Premature end of single
                self.add_tokens_for_single(ignore=True)

            elif tokenum == OP and value == ":":
                # Proper end of single
                self.add_tokens_for_single()

            elif value and self.single.name:
                # Only want to add args after the name for the single has been specified
                self.single.add_to_arg(tokenum, value)

        elif self.after_space or scol == 0 and tokenum == NAME:
            if value in ('describe', 'context'):
                created_group = True

                # add pass to previous group if nothing added between then and now
                if self.groups.empty and not self.groups.root:
                    self.add_tokens_for_pass()

                # Start new group
                self.groups = self.groups.start_group(scol, value)
                self.all_groups.append(self.groups)

            elif value in ('it', 'ignore'):
                self.single = self.groups.start_single(value, scol)

            elif value in ('before_each', 'after_each'):
                setattr(self.groups, "has_%s" % value, True)
                self.add_tokens_for_test_helpers(value)

            else:
                just_append = True
        else:
            # Don't care about it, append!
            just_append = True

        # Found something that isn't whitespace or a new group
        # Hence current group isn't empty !
        if found_content and not created_group:
            self.groups.empty = False

        # Just append if token should be
        if just_append:
            self.result.append([tokenum, value])

    ########################
    ###   UTILITY
    ########################

    def add_tokens(self, tokens):
        """Add tokens to result"""
        self.result.extend([d for d in tokens])

    def reset_indentation(self, amount):
        """Replace previous indentation with desired amount"""
        while self.result and self.result[-1][0] == INDENT:
            self.result.pop()
        self.result.append((INDENT, amount))

    def ignore_token(self):
        """Determine if we should ignore current token"""
        def get_next_ignore(remove=False):
            """Get next ignore from ignore_next and remove from ignore_next"""
            next_ignore = self.ignore_next

            # Just want to return it, don't want to remove yet
            if not remove:
                if type(self.ignore_next) in (list, tuple):
                    next_ignore = self.ignore_next[0]
                return next_ignore

            # Want to remove it from ignore_next
            if type(next_ignore) in (list, tuple) and next_ignore:
                next_ignore = self.ignore_next.pop(0)
            elif not next_ignore:
                self.next_ignore = None
                next_ignore = None
            else:
                self.next_ignore = None

            return next_ignore

        # If we have tokens to be ignored and we're not just inserting till some token
        if not self.insert_till and self.ignore_next:
            # Determine what the next ignore is
            next_ignore = get_next_ignore()
            if next_ignore == (self.current.tokenum, self.current.value):
                # Found the next ignore token, remove it from the stack
                # So that the next ignorable token can be considered
                get_next_ignore(remove=True)
                return True
            else:
                # If not a wildcard, then return now
                if type(next_ignore) is not WildCard:
                    return False

                # Go through tokens untill we find one that isn't a wildcard
                while type(next_ignore) == WildCard:
                    next_ignore = get_next_ignore(remove=True)

                # If the next token is next ignore then we're done here!
                if next_ignore == (self.current.tokenum, self.current.value):
                    get_next_ignore(remove=True)
                    return True
                else:
                    # If there is another token to ignore, then consider the wildcard
                    # And keep inserting till we reach this next ignorable token
                    if next_ignore:
                        self.insert_till = next_ignore
                    return False

    def wrapped_setups(self):
        """Create tokens for Described.setup = noy_wrap_setup(Described, Described.setup) for setup/teardown"""
        lst = []
        for group in self.all_groups:
            if not group.root:
                if group.has_after_each:
                    lst.extend(self.tokens.wrap_after_each(group.kls_name))

                if group.has_before_each:
                    lst.extend(self.tokens.wrap_before_each(group.kls_name))

        if lst:
            indentation_reset = [
                  (NEWLINE, '\n')
                , (INDENT, '')
                ]
            lst = indentation_reset + lst

        return lst

    def make_method_names(self):
        """Create tokens for setting __testname__ on functions"""
        lst = []
        for group in self.all_groups:
            for single in group.singles:
                name, english = single.name, single.english
                if english[1:-1] != name.replace('_', ' '):
                    lst.extend(self.tokens.make_name_modifier(not group.root, single.identifier, english))
        return lst

    def make_describe_attrs(self):
        """Create tokens for setting is_noy_spec on describes"""
        lst = []
        if self.all_groups:
            lst.append((NEWLINE, '\n'))
            lst.append((INDENT, ''))

            for group in self.all_groups:
                if group.name:
                    lst.extend(self.tokens.make_describe_attr(group.kls_name))

        return lst

    def forced_insert(self):
        """
            Insert tokens if self.insert_till hasn't been reached yet
            Will respect self.inserted_line and make sure token is inserted before it
            Returns True if it appends anything or if it reached the insert_till token
        """
        # If we have any tokens we are waiting for
        if self.insert_till:
            # Determine where to append this token
            append_at = -1
            if self.inserted_line:
                append_at = -self.inserted_line+1

            # Reset insert_till if we found it
            if self.current.tokenum == self.insert_till[0] and self.current.value == self.insert_till[1]:
                self.insert_till = None
            else:
                # Adjust self.adjust_indent_at to take into account the new token
                for index, value in enumerate(self.adjust_indent_at):
                    if value < len(self.result) - append_at:
                        self.adjust_indent_at[index] = value + 1

                # Insert the new token
                self.result.insert(append_at, (self.current.tokenum, self.current.value))

            # We appended the token
            return True

    ########################
    ###   ADD TOKENS
    ########################

    def add_tokens_for_pass(self):
        """Add tokens for a pass to result"""
        # Make sure pass not added to group again
        self.groups.empty = False

        # Remove existing newline/indentation
        while self.result[-1][0] in (INDENT, NEWLINE):
            self.result.pop()

        # Add pass and indentation
        self.add_tokens(
            [ (NAME, 'pass')
            , (NEWLINE, '\n')
            , (INDENT, self.indent_type * self.current.scol)
            ]
        )

    def add_tokens_for_test_helpers(self, value):
        """Add setup/teardown function to group"""
        # Add tokens for this block
        tokens = getattr(self.tokens, value)
        self.result.extend(tokens)

        # Add super call if we're inside a class
        if not self.groups.root and not self.wrapped_setup:
            # We need to adjust the indent before the super call later on
            self.adjust_indent_at.append(len(self.result) + 2)

            # Add tokens for super call
            tokens_for_super = self.tokens.make_super(self.indent_type * self.current.scol, self.groups.kls_name, value)
            self.result.extend(tokens_for_super)

            # Tell the machine we inserted a line
            self.inserted_line = len(tokens_for_super)

            # Make sure colon and newline are ignored
            # Already added as part of making super
            self.ignore_next = [(OP, ':'), WildCard(), (NEWLINE, '\n')]

    def add_tokens_for_group(self, with_pass=False):
        """Add the tokens for the group signature"""
        kls = self.groups.super_kls
        name = self.groups.kls_name

        # Reset indentation to beginning and add signature
        self.reset_indentation('')
        self.result.extend(self.tokens.make_describe(kls, name))

        # Add pass if necessary
        if with_pass:
            self.add_tokens_for_pass()

        self.groups.finish_signature()

    def add_tokens_for_single(self, ignore=False):
        """Add the tokens for the single signature"""
        args = self.single.args
        name = self.single.python_name

        # Reset indentation to proper amount and add signature
        self.reset_indentation(self.indent_type * self.single.indent)
        self.result.extend(self.tokens.make_single(name, args))

        # Add skip if necessary
        if ignore:
            self.single.skipped = True
            self.result.extend(self.tokens.test_skip)

        self.groups.finish_signature()

    def finish_hanging(self):
        """Add tokens for hanging singature if any"""
        if self.groups.starting_signature:
            if self.groups.starting_group:
                self.add_tokens_for_group(with_pass=True)

            elif self.groups.starting_single:
                self.add_tokens_for_single(ignore=True)

    ########################
    ###   DETERMINE INFORMATION
    ########################

    def determine_if_whitespace(self):
        """
            Set is_space if current token is whitespace
            Is space if value is:
             * Newline
             * Empty String
             * Something that matches regexes['whitespace']
        """
        value = self.current.value

        if value == '\n':
            self.is_space = True
        else:
            self.is_space = False
            if (value == '' or regexes['whitespace'].match(value)):
                self.is_space = True

    def determine_inside_container(self):
        """
            Set self.in_container if we're inside a container
             * Inside container
             * Current token starts a new container
             * Current token ends all containers
        """
        tokenum, value = self.current.tokenum, self.current.value
        ending_container = False
        starting_container = False

        if tokenum == OP:
            # Record when we're inside a container of some sort (tuple, list, dictionary)
            # So that we can care about that when determining what to do with whitespace
            if value in ['(', '[', '{']:
                # add to the stack because we started a list
                self.containers.append(value)
                starting_container = True

            elif value in [')', ']', '}']:
                # not necessary to check for correctness
                self.containers.pop()
                ending_container = True

        self.just_ended_container = not len(self.containers) and ending_container
        self.just_started_container = len(self.containers) == 1 and starting_container
        self.in_container = len(self.containers) or self.just_ended_container or self.just_started_container

    def determine_indentation(self):
        """Reset indentation for current token and in self.result to be consistent and normalized"""
        # Ensuring NEWLINE tokens are actually specified as such
        if self.current.tokenum != NEWLINE and self.current.value == '\n':
            self.current.tokenum = NEWLINE

        # I want to change dedents into indents, because they seem to screw nesting up
        if self.current.tokenum == DEDENT:
            self.current.tokenum, self.current.value = self.convert_dedent()

        if self.after_space and not self.is_space and (not self.in_container or self.just_started_container):
            # Record current indentation level
            if not self.indent_amounts or self.current.scol > self.indent_amounts[-1]:
                self.indent_amounts.append(self.current.scol)

            # Adjust indent as necessary
            while self.adjust_indent_at:
                self.result[self.adjust_indent_at.pop()] = (INDENT, self.indent_type * (self.current.scol - self.groups.level))

        # Roll back groups as necessary
        if not self.is_space and not self.in_container:
            while not self.groups.root and self.groups.level >= self.current.scol:
                self.finish_hanging()
                self.groups = self.groups.parent

        # Reset indentation to deal with nesting
        if self.current.tokenum == INDENT and not self.groups.root:
           self.current.value = self.current.value[self.groups.level:]

    def convert_dedent(self):
        """Convert a dedent into an indent"""
        # Dedent means go back to last indentation
        if self.indent_amounts:
            self.indent_amounts.pop()

        # Change the token
        tokenum = INDENT

        # Get last indent amount
        last_indent = 0
        if self.indent_amounts:
            last_indent = self.indent_amounts[-1]

        # Make sure we don't have multiple indents in a row
        while self.result[-1][0] == INDENT:
            self.result.pop()

        value = self.indent_type * last_indent
        return tokenum, value




from noseOfYeti.tokeniser.support import noy_wrap_setUp, noy_wrap_tearDown
from should_dsl import should

# Silencing code checker about should_dsl matchers
be = None
equal_to = None

class One(object):
    def setUp(self):
        self.blah = [1]

    def tearDown(self):
        self.meh = [4]

class Two(One):
    def setUp(self):
        self.blah.append(2)

    def tearDown(self):
        self.meh.append(5)

class Three(Two):
    def setUp(self):
        self.blah.append(3)

    def tearDown(self):
        self.meh.append(6)

One.setUp = noy_wrap_setUp(One, One.setUp)
Two.setUp = noy_wrap_setUp(Two, Two.setUp)
Three.setUp = noy_wrap_setUp(Three, Three.setUp)

One.tearDown = noy_wrap_tearDown(One, One.tearDown)
Two.tearDown = noy_wrap_tearDown(Two, Two.tearDown)
Three.tearDown = noy_wrap_tearDown(Three, Three.tearDown)

class TestWrapWorks(object):
    def test_it_ensures_setUp_from_super_classes_get_called(self):
        three = Three()
        hasattr(three, "blah") |should| be(False)
        three.setUp()
        three.blah |should| equal_to([1, 2, 3])

    def test_it_ensures_tearDown_from_super_classes_get_called(self):
        three = Three()
        hasattr(three, "meh") |should| be(False)
        three.tearDown()
        three.meh |should| equal_to([4, 5, 6])




from noseOfYeti.tokeniser import Tokeniser
from should_dsl import should

import six

# Silencing code checker about should_dsl matchers
result_in = None

func_accessor = ""
if six.PY2:
    func_accessor = "__func__ ."

class Test_Tokeniser_Complex(object):
    def setUp(self):
        self.toka = Tokeniser()
        self.tokb = Tokeniser(default_kls = 'other')

        ###   SMALL EXAMPLE

        self.small_example = [
        '''
        describe "This":
            before_each:
                self.x = 5
            describe "That":
                before_each:
                    self.y = 6
                describe "Meh":
                    after_each:
                        self.y = None
            describe "Blah":pass
        describe "Another":
            before_each:
                self.z = 8 '''
        ,
        '''
        class TestThis (%(o)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestThis ,self ));self .x =5
        class TestThis_That (TestThis ):
            def setUp (self ):
                noy_sup_setUp (super (TestThis_That ,self ));self .y =6
        class TestThis_That_Meh (TestThis_That ):
            def tearDown (self ):
                noy_sup_tearDown (super (TestThis_That_Meh ,self ));self .y =None
        class TestThis_Blah (TestThis ):pass
        class TestAnother (%(o)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestAnother ,self ));self .z =8

        TestThis .is_noy_spec =True
        TestThis_That .is_noy_spec =True
        TestThis_That_Meh .is_noy_spec =True
        TestThis_Blah .is_noy_spec =True
        TestAnother .is_noy_spec =True
        '''
        ]

        ###   BIG EXAMPLE

        self.big_example = [
        '''
        describe "This":
            before_each:
                self.x = 5
            it 'should':
                if x:
                    pass
                else:
                    x += 9
            describe "That":
                before_each:
                    self.y = 6
                describe "Meh":
                    after_each:
                        self.y = None
                    it "should set __testname__ for non alpha names ' $^":
                        pass
                    it 'should':
                        if y:
                            pass
                        else:
                            pass
                    it 'should have args', arg1, arg2:
                        blah |should| be_good()
            describe "Blah":pass
        ignore "root level $pecial-method*+"
        describe "Another":
            before_each:
                self.z = 8
            it 'should':
                if z:
                    if u:
                        print "hello \
                            there"
                    else:
                        print "no"
                else:
                    pass
        '''
        ,
        '''
        class TestThis (%(o)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestThis ,self ));self .x =5
            def test_should (self ):
                if x :
                    pass
                else :
                    x +=9
        class TestThis_That (TestThis ):
            def setUp (self ):
                noy_sup_setUp (super (TestThis_That ,self ));self .y =6
        class TestThis_That_Meh (TestThis_That ):
            def tearDown (self ):
                noy_sup_tearDown (super (TestThis_That_Meh ,self ));self .y =None
            def test_should_set_testname_for_non_alpha_names (self ):
                pass
            def test_should (self ):
                if y :
                    pass
                else :
                    pass
            def test_should_have_args (self ,arg1 ,arg2 ):
                blah |should |be_good ()
        class TestThis_Blah (TestThis ):pass
        def ignore__root_level_pecial_method ():raise nose.SkipTest
        class TestAnother (%(o)s ):
            def setUp (self ):
                noy_sup_setUp (super (TestAnother ,self ));self .z =8
            def test_should (self ):
                if z :
                    if u :
                        print "hello \
                            there"
                    else :
                        print "no"
                else :
                    pass

        TestThis .is_noy_spec =True
        TestThis_That .is_noy_spec =True
        TestThis_That_Meh .is_noy_spec =True
        TestThis_Blah .is_noy_spec =True
        TestAnother .is_noy_spec =True
        ignore__root_level_pecial_method .__testname__ ="root level $pecial-method*+"
        TestThis_That_Meh .test_should_set_testname_for_non_alpha_names .{func_accessor}__testname__ ="should set __testname__ for non alpha names ' $^"
        '''.format(func_accessor=func_accessor)
        ]

    ###   TESTS

    def test_works_with_space(self):
        test, desired = self.small_example
        (self.toka, test) |should| result_in(desired % {'o' : 'object'})
        (self.tokb, test) |should| result_in(desired % {'o' : 'other'})

    def test_works_with_tabs(self):
        test, desired = [d.replace('    ', '\t') for d in self.small_example]
        (self.toka, test) |should| result_in(desired % {'o' : 'object'})
        (self.tokb, test) |should| result_in(desired % {'o' : 'other'})

    def test_keeps_good_indentation_in_body_with_spaces(self):
        test, desired = self.big_example
        (self.toka, test) |should| result_in(desired % {'o' : 'object'})
        (self.tokb, test) |should| result_in(desired % {'o' : 'other'})

    def test_keeps_good_indentation_in_body_with_tabs(self):
        test, desired = [d.replace('    ', '\t') for d in self.big_example]
        (self.toka, test) |should| result_in(desired % {'o' : 'object'})
        (self.tokb, test) |should| result_in(desired % {'o' : 'other'})




from setuptools import setup
from textwrap import dedent

setup(
       name = 'noseOfYeti'
     , version = "1.6"
     , classifiers =
       [ 'Intended Audience :: Developers'
       , 'Programming Language :: Python'
       , 'Topic :: Software Development :: Documentation'
       , 'Topic :: Software Development :: Testing'
       ]

     , author = 'Stephen Moore'
     , license = 'GPL'
     , keywords = 'bdd rspec spec'
     , author_email = 'delfick755@gmail.com'

     , url = "https://github.com/delfick/nose-of-yeti"
     , description = 'Nose plugin providing BDD dsl for python'
     , long_description = dedent("""
          Plugin for nose, inspired by http://github.com/fmeyer/yeti, which uses a codec style
          to provide an RSpec style BDD dsl for python tests
       """)

     , extras_require =
       { 'tests':
         [ 'nose'
         , 'fudge'
         , 'should-dsl'
         ]
       , 'docs':
         [ 'nose'
         , 'fudge'
         , 'sphinx'
         , 'pinocchio'
         , 'should_dsl'
         ]
       }

     , install_requires =
       [ 'six'
       ]

     , packages =
       [ 'noseOfYeti'
       , 'noseOfYeti.specs'
       , 'noseOfYeti.plugins'
       , 'noseOfYeti.tokeniser'
       , 'noseOfYeti.plugins.support'
       ]

     , entry_points =
       { 'nose.plugins':
         [ 'noseOfYeti = noseOfYeti.plugins.nosetests:Plugin'
         ]
       }
     )




from contextlib import contextmanager
from textwrap import dedent
import tempfile
import shutil
import os

@contextmanager
def a_temp_file(contents):
    tempfle = None
    try:
        tempfle = tempfile.NamedTemporaryFile(delete=False)
        with open(tempfle.name, 'w') as fle:
            fle.write(dedent(contents))

        yield tempfle.name
    finally:
        if tempfle:
            if os.path.exists(tempfle.name):
                os.remove(tempfle.name)

@contextmanager
def a_temp_dir():
    tempdir = None
    try:
        tempdir = tempfile.mkdtemp()
        yield tempdir
    finally:
        if tempdir:
            if os.path.exists(tempdir):
                shutil.rmtree(tempdir)




from should_dsl import should
import fudge

from noseOfYeti.tokeniser.config import Default, ConfigUtil, MissingConfigFile
from .helpers import a_temp_file

# Silencing code checker about should_dsl matchers
be = None
throw = None
equal_to = None

def test_config_util_initialisation():
    '''Test ConfigUtil initialises properly'''
    template = fudge.Fake("template")
    config_util = ConfigUtil(template)

    config_util.values |should| equal_to({})
    config_util.template |should| be(template)

def test_config_util_normalising_keys():
    '''Test ConfigUtil knows how to normalise a key'''
    config_util = ConfigUtil()
    config_util.normalise_key("blah") |should| equal_to("blah")
    config_util.normalise_key("blah_stuff-things") |should| equal_to("blah_stuff_things")
    config_util.normalise_key("blah-stuff-things") |should| equal_to("blah_stuff_things")

class Test_ConfigUtil_FindingValue(object):
    '''Make sure we can find a value'''
    def test_it_raises_attributeError_when_key_doesnt_exist(self):
        config_util = ConfigUtil()
        config_util.values = {"a":1}

        config_util.find_value("a") |should| be(1)

        finder = lambda : config_util.find_value("b")
        finder |should| throw(AttributeError, "Config has no value for b")

    def test_it_returns_actual_value_if_finds_default(self):
        config_util = ConfigUtil()
        config_util.values = {"a":Default(21)}
        config_util.find_value("a") |should| be(21)

    def test_it_returns_value_as_is(self):
        val = fudge.Fake("val")
        config_util = ConfigUtil()
        config_util.values = {"c":val}
        config_util.find_value("c") |should| be(val)

class Test_ConfigUtil_UsingOptions(object):
    '''Make sure it knows how to use options'''
    def setup(self):
        self.config_util = ConfigUtil()
        self.config_util.values |should| equal_to({})

    def test_it_puts_options_in_values(self):
        self.config_util.use_options({"a":1, "b":2})
        self.config_util.values |should| equal_to({"a":1, "b":2})

    def test_it_overrides_all_values(self):
        c_val = Default(3)
        self.config_util.use_options({"a":1, "b":2, "c":c_val})
        self.config_util.values |should| equal_to({"a":1, "b":2, "c":c_val})

        self.config_util.use_options({"b":4, "c":5, "d":6})
        self.config_util.values |should| equal_to({"a":1, "b":4, "c":5, "d":6})

    def test_it_normalises_keys(self):
        self.config_util.use_options({"a-b":1, "c_d":2, "e-f_g-h":3})
        self.config_util.values |should| equal_to({"a_b":1, "c_d":2, "e_f_g_h":3})

    def test_it_uses_extractor_if_provided(self):
        template = fudge.Fake("template")
        def extractor(templ, values):
            templ |should| be(template)
            for key, val in values.items():
                yield "e-{}".format(key), val + 1

        self.config_util.template = template
        self.config_util.use_options({"a":1, "b":2, "c":3}, extractor=extractor)
        self.config_util.values |should| equal_to({"e_a":2, "e_b":3, "e_c":4})

    def test_it_works_with_list_of_tuples(self):
        c_val = Default(3)
        self.config_util.use_options([("a",1), ("b",2), ("c",c_val)])
        self.config_util.values |should| equal_to({"a":1, "b":2, "c":c_val})

    def test_it_doesnt_complain_if_extractor_returns_none(self):
        '''It's valid for the extractor to say there are no values'''
        template = fudge.Fake("template")
        def extractor(templ, values):
            templ |should| be(template)
            return None

        self.config_util.template = template
        self.config_util.use_options({"a":1, "b":2, "c":3}, extractor=extractor)
        self.config_util.values |should| equal_to({})

class Test_ConfigUtil_FindingConfigFile(object):
    '''Test that the ConfigUtil class knows corner cases of finding a config file'''
    def setUp(self):
        self.config_util = ConfigUtil()
        self.fullpath = fudge.Fake('fullpath')
        self.filename = fudge.Fake('filename')

    @fudge.patch('os.path.exists', 'os.path.abspath')
    def test_it_complains_if_config_file_doesnt_exist(self, fake_exists, fake_abspath):
        self.config_util.values['config_file'] = self.filename
        fake_abspath.expects_call().with_args(self.filename).returns(self.fullpath)
        fake_exists.expects_call().with_args(self.fullpath).returns(False)

        finder = lambda : self.config_util.find_config_file()
        finder |should| throw(MissingConfigFile, "Config file doesn't exist at fake:fullpath")

    @fudge.patch('os.path.exists', 'os.path.abspath')
    def test_it_defaults_to_noyjson(self, fake_exists, fake_abspath):
        fake_abspath.expects_call().with_args('noy.json').returns(self.fullpath)
        fake_exists.expects_call().with_args(self.fullpath).returns(True)
        self.config_util.find_config_file() |should| equal_to(self.fullpath)

    @fudge.patch('os.path.exists', 'os.path.abspath')
    def test_it_doesnt_care_if_default_doesnt_exist(self, fake_exists, fake_abspath):
        fake_abspath.expects_call().with_args('noy.json').returns(self.fullpath)
        fake_exists.expects_call().with_args(self.fullpath).returns(False)
        self.config_util.find_config_file() |should| equal_to(None)

    @fudge.patch('os.path.exists', 'os.path.abspath')
    def test_it_doesnt_care_if_provided_default_doesnt_exist(self, fake_exists, fake_abspath):
        self.config_util.values['config_file'] = Default(self.filename)
        fake_abspath.expects_call().with_args(self.filename).returns(self.fullpath)
        fake_exists.expects_call().with_args(self.fullpath).returns(False)
        self.config_util.find_config_file() |should| equal_to(None)

class Test_ConfigUtil_ApplyingConfigFile(object):
    '''Make sure it knows what values from the config file to apply'''
    def test_it_sets_options_from_an_json_file(self):
        contents = """
            { "a" : "stuff"
            , "b" : "things"
            }
        """
        with a_temp_file(contents) as filename:
            config_util = ConfigUtil()
            config_util.values |should| equal_to({})
            config_util.apply_config_file(filename)
            config_util.values |should| equal_to({"a":"stuff", "b":"things"})

    def test_it_doesnt_override_existing_non_default_values(self):
        contents = """
            { "a" : "stuff"
            , "b" : "things"
            , "c" : "and"
            , "d" : "shells"
            , "e-f" : "blah"
            , "g-h" : "other"
            }
        """
        with a_temp_file(contents) as filename:
            config_util = ConfigUtil()
            c_val = Default(21)
            config_util.use_options({'b':21, 'c':c_val, 'd':None, "g-h":"meh"})
            config_util.values |should| equal_to({'b':21, 'c':c_val, 'd':None, "g_h":"meh"})

            config_util.apply_config_file(filename)
            config_util.values |should| equal_to({"a":"stuff", "b":21, "c":"and", 'd':None, "e_f":"blah", "g_h":"meh"})

class Test_ConfigUtil_UsingConfigFile(object):
    '''Make sure it knows how to find and apply a config file'''
    @fudge.test
    def test_it_doesnt_care_if_no_config_file(self):
        config_util = ConfigUtil()
        fake_find_config_file = fudge.Fake("find_config_file").expects_call().returns(None)
        fake_apply_config_file = fudge.Fake("apply_config_file")

        with fudge.patched_context(config_util, 'find_config_file', fake_find_config_file):
            with fudge.patched_context(config_util, 'apply_config_file', fake_apply_config_file):
                config_util.use_config_file()

    @fudge.test
    def test_it_applies_config_file_if_one_is_found(self):
        config_util = ConfigUtil()
        config_file = fudge.Fake("config_file")
        fake_find_config_file = fudge.Fake("find_config_file").expects_call().returns(config_file)
        fake_apply_config_file = fudge.Fake("apply_config_file").expects_call().with_args(config_file)

        with fudge.patched_context(config_util, 'find_config_file', fake_find_config_file):
            with fudge.patched_context(config_util, 'apply_config_file', fake_apply_config_file):
                config_util.use_config_file()




from noseOfYeti.tokeniser.spec_codec import register_from_options
from noseOfYeti.plugins.support.spec_options import spec_options
from noseOfYeti.plugins.support.test_chooser import TestChooser
from noseOfYeti.tokeniser.config import Default

from nose2.events import Plugin

import logging
import sys
import os

log = logging.getLogger('nose2.plugins.noseOfYeti')

def extract_options(template, options):
    for option, val in template.items():
        yield option, options.get(option)

class NoseOfYetiPlugin(Plugin):
    configSection = 'noseOfYeti'

    def __init__(self):
        self.test_chooser = TestChooser()
        default_ignore_kls = []
        if "NOSE_NOY_IGNORE_KLS" in os.environ:
            default_ignore_kls.extend(os.environ["NOSE_NOY_IGNORE_KLS"].split(","))
        self.ignore_kls = self.config.as_list('ignore-kls', default=default_ignore_kls)
        self.always_on = self.config.as_bool('always-on', default=False)

        parser_options = ['default', 'action', 'dest', 'help']
        for option, attributes in spec_options.items():
            action = attributes.get("action", "store")
            default = attributes.get("default")

            if action == "store_true":
                action = "as_bool"
                default = False
            elif action == "store_false":
                action = "as_bool"
                default = True
            elif action == "append":
                action = "as_list"
                default = attributes.get("default", [])
            elif action == "store":
                action = "as_str"
                default = attributes.get("default")

            if callable(default):
                default = default(os.environ)
            setattr(self, option.replace("-", "_"), getattr(self.config, action)(option, default=Default(default)))

    def handleFile(self, event):
        if not getattr(self, "_configured", False):
            self._configured = True
            self.enable()

    def loadTestsFromModule(self, event):
        self.test_chooser.new_module()
        if hasattr(event.module, "__package__"):
            pkg = sys.modules.get(event.module.__package__)
            if getattr(pkg, "__test__", None) is False:
                event.handled = True

    def getTestCaseNames(self, event):
        names = filter(event.isTestMethod, dir(event.testCase))
        methods = [(name, getattr(event.testCase, name)) for name in names]
        event.handled = True
        return [name for name, method in methods if self.test_chooser.consider(method, self.ignore_kls) is not False]

    def enable(self):
        self.done = {}
        options = dict((option, getattr(self, option.replace('-', '_'))) for option in spec_options)
        register_from_options(options, spec_options, extractor=extract_options)

